import importlib
import sys
from typing import List, Type

from heare.config import SettingsDefinition, ListSetting

from heare.client.events.ws import HeareClient, HeareClientSettings
from heare.logging import get_logger
from heare.skills import Skill

logger = get_logger(__name__)


class ContainerSettings(SettingsDefinition):
    skill = ListSetting(str)


class HeareContainer(object):
    def __init__(self,
                 skill_classes: List[Type],
                 heare_client: HeareClient,
                 args: List[str]):
        self.skill_classes: List[Type] = skill_classes
        self.heare_client: HeareClient = heare_client
        self.args: List[str] = args

    def bootstrap(self):
        def wire_skills(*args):
            for klass in self.skill_classes:
                try:
                    skill: Skill = klass(event_source=self.heare_client)
                    config = None
                    if hasattr(klass, '__SETTINGS__'):
                        config = getattr(klass, '__SETTINGS__').load(args=args)
                    skill.load(settings=config)
                except Exception as e:
                    logger.exception(f"Error instantiating skill {klass}")

        self.heare_client.once('open', wire_skills)


def hydrate_skill_classes(container_settings: ContainerSettings) -> List[Type]:
    skill_classes = []
    for skill_str in container_settings.skill.get():
        try:
            module_name, klass_name = skill_str.rsplit('.', 1)
            mod = importlib.import_module(module_name)
            skill_classes.append(
                getattr(mod, klass_name)
            )
        except Exception as e:
            logger.exception(f"Error loading skill: {skill_str}.")

    return skill_classes


def main(args):
    container_settings = ContainerSettings.load(args=args)
    heare_client_settings = HeareClientSettings.load(args=args)
    client = HeareClient(config=heare_client_settings)

    skill_classes = hydrate_skill_classes(container_settings)

    container = HeareContainer(skill_classes, client, args)
    container.bootstrap()
    client.run_forever()


if __name__ == '__main__':
    main(sys.argv[1:])
